#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
""" Unit tests for Capability Service """

# pylint: disable=C0116, C0301, E0401, R0201
from datetime import datetime
from unittest.mock import patch

import pytest

from workspaces.capability.enums import CapabilityVersionState
from workspaces.capability.schema import (
    Capability,
    CapabilityExecution,
    CapabilityRequest,
    CapabilityVersion,
)
from workspaces.capability.services.capability_info import CapabilityInfo
from workspaces.capability.services.capability_service import CapabilityService
from workspaces.notification.services.interfaces import NotificationServiceIF


@pytest.mark.usefixtures("mock_capability_service")
class TestCapabilityService:
    """Tests for CapabilityService methods"""

    def test_on_ingestion_complete(
        self,
        mock_capability_service: CapabilityService,
        mock_capability_info: CapabilityInfo,
    ):
        """
        Does on_ingestion_complete() save the request and send the "ingestion complete" message?

        :param mock_capability_service: mocked CapabilityService
        :param mock_capability_info: mocked CapabilityInfo
        :return:
        """
        fake_capability = Capability()
        fake_capability.name = "null"

        fake_capability_request = CapabilityRequest()
        fake_capability_request.id = 1
        fake_capability_request.capability = fake_capability
        fake_capability_request.capability_name = fake_capability.name

        fake_exec = CapabilityExecution()
        fake_exec.created_at = fake_exec.updated_at = datetime.now()
        fake_exec.capability = fake_capability
        fake_exec.capability_request_id = fake_capability_request.id
        fake_exec.capability_name = fake_capability.name
        fake_exec.version_number = 1
        fake_exec.state_name = CapabilityVersionState.Complete.name
        fake_exec.current_workflow_request_id = 1

        fake_v1 = CapabilityVersion()
        fake_v1.version_number = 1
        fake_v1.capability = fake_capability
        fake_v1.capability_name = fake_capability.name
        fake_v1.capability_request_id = fake_capability_request.id
        fake_v1.request = fake_capability_request
        fake_v1.capability_request = fake_capability_request
        fake_v1.parameters = {"user_email": "im_a_little_teapot@rabbithole.org"}
        fake_v1.executions = [fake_exec]

        fake_subject = {
            "type": "WorkflowRequest",
            "workflow_request_id": -1,
            "execution_wf_id": "-1",
        }
        fake_ingestion_complete_msg = {
            "subject": fake_subject,
            "service": "workflow",
            "routing_key": "workflow",
            "type": "ingestion-complete",
        }

        with patch(
            "workspaces.capability.services.capability_info.CapabilityInfo.lookup_capability_request",
            return_value=fake_capability_request,
        ):
            with patch(
                "workspaces.capability.services.capability_info.CapabilityInfo.lookup_execution_by_workflow_request_id",
                return_value=fake_exec,
            ):
                with patch("workspaces.capability.services.capability_info.CapabilityInfo.save_entity"):
                    with patch(
                        "workspaces.notification.services.notification_service.NotificationServiceIF"
                        ".notify_ingestion_complete"
                    ) as mock_notify:
                        notify_old_call_count = NotificationServiceIF.notify_ingestion_complete.call_count
                        mock_capability_service.on_ingestion_complete(**fake_ingestion_complete_msg)
                        assert mock_notify.call_count == notify_old_call_count + 1
                        (request,) = mock_capability_info.save_entity.call_args.args
                        assert request.ingested
                        assert request.sealed

    def test_on_carta_ready(
        self,
        mock_capability_info: CapabilityInfo,
        mock_capability_service: CapabilityService,
    ):
        """
        Does on_carta_ready() save the CARTA-ready request version?

        :param mock_capability_info: passes for a CapabilityInfo
        :param mock_capability_service: thinks it's a CapabilityService
        :return:
        """
        fake_capability = Capability()
        fake_capability.name = "null"

        fake_capability_request = CapabilityRequest()
        fake_capability_request.id = 1
        fake_capability_request.capability = fake_capability
        fake_capability_request.capability_name = fake_capability.name

        fake_exec = CapabilityExecution()
        fake_exec.created_at = fake_exec.updated_at = datetime.now()
        fake_exec.capability = fake_capability
        fake_exec.capability_request_id = fake_capability_request.id
        fake_exec.capability_name = fake_capability.name
        fake_exec.version_number = 1
        fake_exec.state_name = CapabilityVersionState.Complete.name

        fake_v1 = CapabilityVersion()
        fake_v1.version_number = 1
        fake_v1.capability = fake_capability
        fake_v1.capability_name = fake_capability.name
        fake_v1.capability_request_id = fake_capability_request.id
        fake_v1.request = fake_capability_request
        fake_v1.capability_request = fake_capability_request
        fake_v1.parameters = {"user_email": "nobody@nowhere.edu"}
        fake_v1.executions = [fake_exec]

        wf_request_id = -2
        carta_url = "descartes_image_carta_url"
        fake_carta_ready_msg = {
            "service": "capability",
            "routing_key": "capability",
            "carta_url": carta_url,
            "subject": {"workflow_request_id": wf_request_id, "argument": {"parent_wf_request_id": -1}},
            "type": "carta-ready",
        }
        with patch(
            "workspaces.capability.services.capability_info.CapabilityInfo.lookup_execution_by_workflow_request_id",
            return_value=fake_exec,
        ):
            with patch("workspaces.capability.services.capability_info.CapabilityInfo.save_entity"):
                save_entity_old_call_count = mock_capability_info.save_entity.call_count
                mock_capability_service.on_carta_ready(**fake_carta_ready_msg)
                assert mock_capability_info.save_entity.call_count == save_entity_old_call_count + 1

                (request_version,) = mock_capability_info.save_entity.call_args.args
                assert request_version.version_number > 0
                assert request_version.workflow_metadata["carta_url"] == carta_url
