#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
""" This is the execution manager. """

# pylint: disable=C0301, E0401, W1203

import logging
from typing import Dict, List, Optional

import transaction
from messaging.messenger import MessageSender
from messaging.router import Router, on_message

from workspaces.capability.enums import CapabilityRequestState, QueueState
from workspaces.capability.message_architect import CapabilityMessageArchitect
from workspaces.capability.schema import CapabilityExecution
from workspaces.capability.services.capability_info import CapabilityInfo
from workspaces.capability.services.interfaces import ExecutionManagerIF
from workspaces.capability.services.queue_manager import CapabilityQueueManager
from workspaces.capability.services.restriction_engine import Restrictions
from workspaces.notification.services.interfaces import NotificationServiceIF
from workspaces.system.schema import AbstractFile
from workspaces.workflow.services.interfaces import WorkflowServiceIF

logger = logging.getLogger(__name__)


class ExecutionManager(ExecutionManagerIF):
    """
    Class that manages the capability execution system. Its duties include:
    - Creating new executions
    - Managing existing execution state
    - Handling capability events and updating the state of the system accordingly
    """

    def __init__(
        self,
        capability_info: CapabilityInfo,
        workflow_service: WorkflowServiceIF,
        router: Router,
        messenger: MessageSender,
        notifier: NotificationServiceIF,
    ):
        self.capability_info = capability_info
        self.workflow_service = workflow_service
        self.messenger = messenger
        self.queue_manager = CapabilityQueueManager(self.capability_info, self)
        self._notifier = notifier
        self.router = router
        self.router.register(self)

    @property
    def notifier(self) -> NotificationServiceIF:
        return self._notifier

    def queue_execution(self, execution: CapabilityExecution):
        """
        Queue this execution, or else put it in the queue (if it needs to run a PrepareAndRunWorkflow)

        :param execution:
        """
        # mark this execution as queued
        execution.queue_state = QueueState.Queued.name
        self.capability_info.save_execution(execution)

        # now tell the queue manager to handle it, once we're done with everything else
        transaction.get().addAfterCommitHook(lambda _: transaction.begin() and self.queue_manager.execution_requested())

    def resume_execution(self, execution: CapabilityExecution):
        """
        Called by the queue manager when the execution can resume
        :param execution:  the execution whose time has come
        """
        # Mark this execution as no longer queued
        execution.queue_state = QueueState.Running.name
        self.capability_info.save_execution(execution)
        subject = execution.__json__()

        def send_message(status, execution):
            message = {
                "type": "resume",
                "service": "capability",
                "subject": execution,
            }
            self.messenger.send_message(**message)

        transaction.get().addAfterCommitHook(send_message, (subject,))
        transaction.commit()

    def execute(self, execution: CapabilityExecution) -> CapabilityExecution:
        """
        Run a capability with the requested settings

        :param execution: Capability execution
        :return: Execution of the request
        """
        request = execution.capability_request
        request.state = CapabilityRequestState.Submitted.name
        self.capability_info.save_entity(request)
        return self.start(execution)

    def start(self, execution: CapabilityExecution):
        message = {
            "type": "start",
            "service": "capability",
            "subject": execution.__json__(),
        }
        self.messenger.send_message(**message)
        return execution

    @on_message(service="workflow", type="workflow-complete")
    @on_message(service="workflow", type="workflow-failed")
    def mark_execution_as_not_running(self, **message):
        # As a friendly service, we update the queue state here regardless of whatever other processing occurs
        if execution := self.execution_from_message(**message):
            execution.queue_state = QueueState.NotQueued.name
            self.capability_info.save_execution(execution)
            self.queue_manager.execution_complete()

    @on_message(service="capability")
    @on_message(service="workflow")
    def process_message(self, **message: Dict):
        """
        Advance execution that is waiting to its next state

        :param message: Workflow-execution message
        """
        logger.info(f"In process_message, received message: {message}")

        # we ignore a few types of message that we mostly generate
        if message["type"] in ["capability-complete", "execution-complete", "restart-queue"]:
            return

        try:
            if execution := self.execution_from_message(**message):
                logger.info("Found execution %s", execution.id)
                execution.signal(self, message)
                self.capability_info.save_execution(execution)
                self.check_for_complete(execution)
                transaction.commit()
            else:
                logger.info("Did not find an execution in this message somehow")
        except Exception as e:
            logger.exception("Something went wrong", e)

    def submit_workflow_request(
        self,
        execution_id: int,
        workflow_name: str,
        workflow_args: dict,
        files: List[AbstractFile],
    ):
        """
        Submit a workflow request to the workflow service

        :param execution_id:
        :param workflow_name:
        :param workflow_args:
        :param files:
        """
        logger.info("Submitting workflow request")
        transfer_files_string, transfer_files_list = self._get_transfer_files_properties(files)
        args_list = {**workflow_args, "files_to_transfer": transfer_files_string, "provided_files": transfer_files_list}

        execution = self.capability_info.lookup_execution(execution_id)

        workflow_request = self.capability_info.restriction_engine.trigger(
            execution,
            Restrictions.CREATE_WORKFLOW,
            trigger_args=[self.workflow_service, workflow_name, args_list, files],
        )

        if workflow_request:

            def execute_workflow_request(status, request):
                logger.info("Attempting to execute workflow request %s", request)
                self.workflow_service.execute(request)

            transaction.get().addAfterCommitHook(execute_workflow_request, (workflow_request,))

    def _execution_from_child_message(self, **message: Dict) -> Optional[CapabilityExecution]:
        """
        Extract a CapabilityExecution from a message sent from a child workflow request

        :param message: Message sent from a child workflow request
        :return: CapabilityExecution if one can be located for child request; else None
        """
        msg_subject = message.get("subject")
        parent_wf_request_id = None

        if msg_subject.get("execution_wf_id"):
            parent_wf_request_id = msg_subject.get("execution_wf_id")
        elif wf_arguments := message.get("argument"):
            if wf_arguments.get("parent_wf_request_id"):
                parent_wf_request_id = message["argument"].get("parent_wf_request_id")

        if parent_wf_request_id:
            # Workflow request has a parent_wf_request_id; use that as the request ID instead of the true ID
            logger.info("Trying to find an execution with parent_wf_request_id=%s", parent_wf_request_id)
            return self.capability_info.lookup_execution_by_workflow_request_id(parent_wf_request_id)
        return None

    def _execution_from_qa_message(self, **message: Dict) -> Optional[CapabilityExecution]:
        """
        Extract a CapabilityExecution from a QA message

        :param message: QA message
        :return: CapabilityExecution if one can be located; else None
        """
        msg_subject = message.get("subject")
        version_number = message["capability_version"]
        execution = self.capability_info.lookup_execution_by_workflow_request_id(msg_subject["workflow_request_id"])
        if version_number != execution.version_number:
            logger.error("ERROR: Specified version does not match retrieved execution.")
            return None
        else:
            return execution

    def execution_from_message(self, **message: Dict) -> Optional[CapabilityExecution]:
        """
        Extract a CapabilityExecution from a message

        :param message: Message
        :return: CapabilityExecution if ID is found in message; else None
        """
        msg_subject = message.get("subject")
        msg_type = msg_subject.get("type")

        # first, let's check for a capability execution ID
        if msg_type == "CapabilityExecution":
            # if we found one, we're basically done
            return self.capability_info.lookup_execution(msg_subject["id"])
        elif msg_type == "CapabilityRequest":
            return self.capability_info.lookup_execution(msg_subject["current_execution"]["id"])
        elif message["service"] == "workflow":
            # OK, the message is from the workflow layer
            # Look for a parent_wf_request_id first
            if msg_subject:
                if execution := self._execution_from_child_message(**message):
                    return execution

            # No parent ID found, so simply use the request's own workflow request ID
            logger.info("Trying to find an execution with workflow_request_id=%s", msg_subject["workflow_request_id"])
            if "qa" in msg_type:
                self._execution_from_qa_message(**message)

            return self.capability_info.lookup_execution_by_workflow_request_id(msg_subject["workflow_request_id"])
        logger.error("Error: unable to ascertain the capability execution for message %s", message)
        return None

    def _complete_execution(self, execution: CapabilityExecution) -> CapabilityExecution:
        """
        Perform necessary steps to declare a capability execution complete

        :param execution: Execution to complete
        :return: Execution with updated metadata
        """
        execution.queue_state = QueueState.NotQueued.name

        execution_complete_msg = CapabilityMessageArchitect(execution=execution).compose_message("execution_complete")
        self.messenger.send_message(**execution_complete_msg)

        return execution

    def _fail_execution(self, execution: CapabilityExecution) -> CapabilityExecution:
        """
        Perform necessary steps to declare a capability execution failed

        :param execution: Execution to fail
        :return: Execution with updated metadata
        """
        execution.queue_state = QueueState.NotQueued.name

        execution_complete_msg = CapabilityMessageArchitect(execution=execution).compose_message("execution_failed")
        self.messenger.send_message(**execution_complete_msg)

        return execution

    def _error_execution(self, execution: CapabilityExecution) -> CapabilityExecution:
        """
        Perform necessary steps to declare a capability execution errored

        :param execution: Execution to error
        :return: Execution with updated metadata
        """
        execution.queue_state = QueueState.NotQueued.name

        execution_complete_msg = CapabilityMessageArchitect(execution=execution).compose_message("execution_error")
        self.messenger.send_message(**execution_complete_msg)

        return execution

    def _cancel_execution(self, execution: CapabilityExecution) -> CapabilityExecution:
        """
        Perform steps to finalize a cancelled capability execution


        :param execution: Execution to cancel
        :return: Execution with updated metadata
        """
        execution.queue_state = QueueState.NotQueued.name
        return execution

    def check_for_complete(self, execution: CapabilityExecution):
        if execution.state.final():
            if execution.state == "Complete":
                self._complete_execution(execution)
            elif execution.state == "Failed":
                self._fail_execution(execution)
            elif execution.state == "Error":
                self._error_execution(execution)
            elif execution.state == "Cancelled":
                self._cancel_execution(execution)

    @staticmethod
    def _get_transfer_files_properties(files: List[AbstractFile]):
        name_list = []
        separator = ", ./"
        # needed for template rendering
        if files:
            for file in files:
                name_list.append(file.filename)

            return [separator + separator.join(file.filename for file in files), name_list]
        else:
            return ["", []]
